/*
 *  Copyright (c) 2007-2008 Riccardo Caselli
 *
 *  USB Logitech MX610 Cordless Mouse driver
 *
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb.h>

#define DRIVER_VERSION	"v0.0.3"
#define DRIVER_AUTHOR	"Riccardo Caselli"
#define DRIVER_DESC	"USB Logitech MX610 Cordless Mouse driver"
#define DRIVER_LICENSE	"GPL"

#define GETBIT(v,n)     ((v>>(n))&0x01)
#define SETBIT(v,n)     (v |= (0x01<<(n)))

#ifdef SLAB_ATOMIC
# define ATOMIC SLAB_ATOMIC
#else
# define ATOMIC GFP_ATOMIC
#endif

/* Module properties */

MODULE_AUTHOR ( DRIVER_AUTHOR );
MODULE_DESCRIPTION ( DRIVER_DESC );
MODULE_LICENSE ( DRIVER_LICENSE );

/* Own type */

typedef struct usb_mx610 {

	// Device name

	char name[128];

	// USB interrupt data

	signed char *data;

	char phys[64];

	dma_addr_t data_dma;

	// USB device

	struct usb_device *usbdev;

	// Input device

	struct input_dev *inputdev;

	// USB Request block

	struct urb *urb;

	// Number of openned times

	int open;

} mx610_t;


// Initialize mx610 structure

void mx610_init ( mx610_t *mx610 ) {

	memset(mx610, 0, sizeof(mx610_t));
	mx610->inputdev = NULL;
	mx610->urb = NULL;
	mx610->data = NULL;

}


// Free mx610 buffers

void mx610_free ( mx610_t *mx610 ) {

	if ( mx610->urb )
		usb_free_urb(mx610->urb);

	if ( mx610->data )
		usb_buffer_free(mx610->usbdev,8,mx610->data,mx610->data_dma);

	kfree(mx610);	

}


// Create new mx610 (buffer allocation

mx610_t *mx610_new ( struct usb_device *dev ) {

	mx610_t *mx610;

	// Create object

	if (!(mx610 = kmalloc(sizeof(mx610_t), GFP_KERNEL)))
		return NULL;

	// Initialize

	mx610_init(mx610);

	// Input device

	if ( (mx610->inputdev = input_allocate_device()) == NULL ) {
		mx610_free(mx610);
		return NULL;
	}

	// Create urb handler

	if (!(mx610->urb = usb_alloc_urb(0, GFP_KERNEL))) {
		mx610_free(mx610);
		return NULL;
	}


	// Create data required for urb transfer

	if (!(mx610->data = usb_buffer_alloc(dev,8,ATOMIC,&mx610->data_dma))) {
		mx610_free(mx610);
		return NULL;
	}


	// Set mx610 usb device

	mx610->usbdev = dev;


	return mx610;

}





// Get data from urb and send to input API

void input_send_data ( struct input_dev *dev, char *data ) {
	
	char
		btn = data[0],	// Basic buttons (left, right, middle, side)
		mbtn = data[1],	// Mail and IM buttons
		xbuff = data[2],	// X movement
		xybuff = data[3], // Last 4 bits used for Y movement
		ybuff = data[4],	// First 4 bits used for Y movement
		w = data[5],	// Wheel movement
		hwheel = data[6],	// Horizontal Wheel mapped as 2 button
		volup,
		vold,
		mute,
		email,
		im,
		em3btn;
	
	int i,y=0,x=0;
/*
	int j;
	for (i=0;i<8;i++) {
		for (j=0;j<8;j++) {
			printk("%d",GETBIT(data[i],j));
		}
		printk(" ");
	}
	printk("\n");
*/

	em3btn=GETBIT(btn,0)&GETBIT(btn,1);
	volup=GETBIT(btn,0)&GETBIT(btn,1)&GETBIT(data[3],0)&GETBIT(data[3],3)&GETBIT(data[3],5)&GETBIT(data[3],6)&GETBIT(data[3],7);
	vold=GETBIT(btn,0)&GETBIT(btn,1)&GETBIT(data[3],1)&GETBIT(data[3],3)&GETBIT(data[3],5)&GETBIT(data[3],6)&GETBIT(data[3],7);
	mute=GETBIT(btn,0)&GETBIT(btn,1)&GETBIT(data[3],1)&!GETBIT(data[3],3)&GETBIT(data[3],5)&GETBIT(data[3],6)&GETBIT(data[3],7);
	email=GETBIT(btn,0)&GETBIT(btn,1)&GETBIT(mbtn,1)&GETBIT(mbtn,3)&GETBIT(mbtn,7)&GETBIT(data[2],0);
	im=GETBIT(btn,0)&GETBIT(btn,1)&GETBIT(mbtn,2)&GETBIT(mbtn,3)&GETBIT(mbtn,4)&GETBIT(mbtn,5)&GETBIT(mbtn,7)&GETBIT(data[2],0);

	if (volup||vold||mute||email||im||em3btn) {
		btn=0;
		xbuff=0;
		xybuff=0;
		ybuff=0;
	}
	
	x=xbuff;
	for (i=0;i<4;i++) {
		if (GETBIT(xybuff,i)) {
			SETBIT(x,i+8);
		}
	}
	y=ybuff;
	y=y<<(4);
	xybuff=xybuff>>(4);
	for (i=0;i<4;i++) {
		if (GETBIT(xybuff,i)) {
			SETBIT(y,i);
		}
	}
	
	input_report_key(dev, BTN_LEFT,   	GETBIT(btn,0));
	input_report_key(dev, BTN_RIGHT,  	GETBIT(btn,1));
	input_report_key(dev, BTN_MIDDLE, 	GETBIT(btn,2));
	input_report_key(dev, KEY_FORWARD,	GETBIT(btn,4));
	input_report_key(dev, KEY_BACK,		GETBIT(btn,3));
	input_report_key(dev, KEY_NEXTSONG,	GETBIT(hwheel,0)&!GETBIT(hwheel,1));
	input_report_key(dev, KEY_NEXTSONG,	0);
	input_report_key(dev, KEY_PREVIOUSSONG,	GETBIT(hwheel,0)&GETBIT(hwheel,1));
	input_report_key(dev, KEY_PREVIOUSSONG,	0);

	input_report_key(dev, KEY_VOLUMEUP,	volup);
	input_report_key(dev, KEY_VOLUMEDOWN, vold);
	input_report_key(dev, KEY_MUTE,	mute);
	input_report_key(dev, KEY_EMAIL, email);
	input_report_key(dev, KEY_CHAT,	im);

	input_report_rel(dev, REL_X,     x);
	input_report_rel(dev, REL_Y,     y);
	input_report_rel(dev, REL_WHEEL, w);

}

static void usb_mx610_handle(struct urb *urb) {

	mx610_t *mouse = urb->context;
	signed char *data = mouse->data;
	struct input_dev *inputdev = mouse->inputdev;


	// Check returned status

	if (urb->status) return;


	// Send data to input interface

	input_send_data(inputdev,data);

	input_sync(inputdev);
	usb_submit_urb(urb,ATOMIC);

}

static int usb_mx610_open(struct input_dev *dev) {

	mx610_t *mouse = dev->private;

	if (mouse->open++)
		return 0;

	mouse->urb->dev = mouse->usbdev;

	if (usb_submit_urb(mouse->urb, GFP_KERNEL)) {
		mouse->open--;
		return -EIO;

	}

	return 0;

}

static void usb_mx610_close(struct input_dev *dev) {

	mx610_t *mouse = dev->private;

	if (!--mouse->open)
		usb_kill_urb(mouse->urb);

}

static void input_device_init ( struct input_dev *inputdev, struct usb_interface *intf, void *private, struct usb_device *dev ) {

	char path[64];

	mx610_t *mouse = (mx610_t *) private;

	int
		x,
		keys[]	= { KEY_BACK, KEY_FORWARD, KEY_VOLUMEUP, KEY_VOLUMEDOWN, KEY_MUTE, KEY_EMAIL, KEY_CHAT,
			    KEY_NEXTSONG, KEY_PREVIOUSSONG, 0 };


	// Events

	inputdev->evbit[0] = BIT(EV_KEY) | BIT(EV_REL);

	// Add special keys

	for ( x = 0 ; keys[x] ; x++ )
		set_bit(keys[x],inputdev->keybit);

	// Add basic buttons

	inputdev->keybit[LONG(BTN_MOUSE)] = BIT(BTN_LEFT) | BIT(BTN_RIGHT) | BIT(BTN_MIDDLE) |
					    BIT(BTN_SIDE) | BIT(BTN_EXTRA);

	// Add move mouse movement (X/Y)

	inputdev->relbit[0] = BIT(REL_X) | BIT(REL_Y);

	// Add wheel

	inputdev->relbit[0] |= BIT(REL_WHEEL);



	// Private data structure

	inputdev->private = mouse;

	// Input file operations

	inputdev->open = usb_mx610_open;
	inputdev->close = usb_mx610_close;

	// Device

	inputdev->name = mouse->name;

	usb_make_path(dev,path,64);
	snprintf(mouse->phys,64,"%s/input0",path);

	inputdev->phys = mouse->phys;
	inputdev->id.bustype = BUS_USB;
	inputdev->id.vendor = dev->descriptor.idVendor;
	inputdev->id.product = dev->descriptor.idProduct;
	inputdev->id.version = dev->descriptor.bcdDevice;

}

static int usb_mx610_probe(struct usb_interface *intf, const struct usb_device_id *id) {

	struct usb_device *dev = interface_to_usbdev(intf);
	struct usb_host_interface *interface;


	struct usb_endpoint_descriptor *endpoint;
	mx610_t *mouse;
	int pipe, maxp;
	char *buf;


	// Get mouse endpoint

	interface = intf->cur_altsetting;

	if ( interface->desc.bNumEndpoints != 1 ) return -ENODEV;
	endpoint = &interface->endpoint[0].desc;


	// Check endpoint

	if (!(endpoint->bEndpointAddress & USB_DIR_IN))
		return -ENODEV;

	if ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) != USB_ENDPOINT_XFER_INT)
		return -ENODEV;


	// Create endpoint pipe

	pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
	maxp = usb_maxpacket(dev, pipe, usb_pipeout(pipe));


	// Create mx610 object

	if (!(mouse = mx610_new(dev)))
		return -ENOMEM;

	// Initialize input device

	input_device_init(mouse->inputdev,intf,mouse,dev);


	// Set device name

	if (!(buf = kmalloc(63, GFP_KERNEL))) {
		mx610_free(mouse);
		return -ENOMEM;
	}


	if (dev->descriptor.iManufacturer &&
		usb_string(dev, dev->descriptor.iManufacturer, buf, 63) > 0)
			strcat(mouse->name, buf);

	if (dev->descriptor.iProduct &&
		usb_string(dev, dev->descriptor.iProduct, buf, 63) > 0)
			sprintf(mouse->name, "%s %s", mouse->name, buf);

	if (!strlen(mouse->name))
		sprintf(mouse->name, "mx610_usb.c: Logitech MX610 Mouse on usb%04x:%04x",
			mouse->inputdev->id.vendor, mouse->inputdev->id.product);

	kfree(buf);


	// Initialize interrupt transfer

	usb_fill_int_urb(mouse->urb,dev,pipe,mouse->data,((maxp > 8)?8:maxp),usb_mx610_handle,mouse,endpoint->bInterval);
	mouse->urb->transfer_dma = mouse->data_dma;
	mouse->urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;


	// Register input device

	input_register_device(mouse->inputdev);


	printk(KERN_INFO "mx610_usb.c: %s\n", mouse->name);

	// Set usb handler interface data

	usb_set_intfdata(intf,mouse);


	return 0;

}


static void usb_mx610_disconnect(struct usb_interface *intf) {

	mx610_t *mouse = usb_get_intfdata(intf);

	usb_set_intfdata(intf,NULL);
	if (mouse) {
		usb_kill_urb(mouse->urb);
		input_unregister_device(mouse->inputdev);
		mx610_free(mouse);
	}

}




/* Module structures */

static struct usb_device_id usb_mx610_id_table [] = {
	{ USB_DEVICE(0x46d, 0xc518) },
	{ }
};

MODULE_DEVICE_TABLE (usb, usb_mx610_id_table);

static struct usb_driver usb_mx610_driver = {

	.name		= "mx610_usb",
	.probe		= usb_mx610_probe,
	.disconnect	= usb_mx610_disconnect,
	.id_table	= usb_mx610_id_table

};



/* Module main functions */

static int __init usb_mx610_init(void) {

	int rv;

	// Register usb driver

	rv = usb_register(&usb_mx610_driver);

	info(DRIVER_VERSION ":" DRIVER_DESC);

	return rv;
}

static void __exit usb_mx610_exit(void) {

	usb_deregister(&usb_mx610_driver);

}

// Set

module_init(usb_mx610_init);
module_exit(usb_mx610_exit);
